/*
 * ubus.c: URL filter ubus method
 *
 * Copyright (C) 2024 iopsys Software Solutions AB. All rights reserved.
 *
 * Author: Rahul Thakur <rahul.thakur@iopsys.eu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */

#include <libubus.h>
#include <syslog.h>

#include "filter.h"
#include "dns_cache.h"
#include "utils.h"
#include "ubus.h"

struct ubus_context *ctx = NULL;
static const char *ubus_path;

void print_url_list(struct list_head *lh, struct blob_buf *b)
{
	if (!lh)
		return;

	struct url_list *url;

	list_for_each_entry(url, lh, list) {
		void *table;
		table = blobmsg_open_table(b, "");

		blobmsg_add_string(b, "url", url->url);
		blobmsg_add_u64(b, "hits", url->match_cnt);

		blobmsg_close_table(b, table);
	}
}

static int dump_dns_cache(struct ubus_context *context, struct ubus_object *obj,
			struct ubus_request_data *req, const char *method,
			struct blob_attr *msg)
{
	void *cache_array = NULL, *cache_table = NULL;
	void *ipv4_array = NULL, *ipv4_table = NULL;
	void *ipv6_array = NULL, *ipv6_table = NULL;
	void *cname_array = NULL, *cname_table = NULL;
	struct blob_buf b = {0};
	memset(&b, 0, sizeof(struct blob_buf));
	blob_buf_init(&b, 0);

	cache_array = blobmsg_open_array(&b, "dns_cache");
	for (size_t i = 0; i < dns_cache_cnt; i++) {
		cache_table = blobmsg_open_table(&b, "");

		dns_cache *cache = &global_dns_cache[i];
		blobmsg_add_string(&b, "url", cache->url);
		blobmsg_add_u64(&b, "hits", (uint64_t)cache->match_cnt);

		if (cache->ip_addr_cnt > 0) {
			ipv4_array = blobmsg_open_array(&b, "ipv4_addr");
			for (size_t j = 0; j < cache->ip_addr_cnt; j++) {
				ipv4_table = blobmsg_open_table(&b, "");
				blobmsg_add_string(&b, "addr", cache->ip_addr[j]);
				blobmsg_close_table(&b, ipv4_table);
			}
			blobmsg_close_array(&b, ipv4_array);
		}

		if (cache->ipv6_addr_cnt > 0) {
			ipv6_array = blobmsg_open_array(&b, "ipv6_addr");
			for (size_t j = 0; j < cache->ipv6_addr_cnt; j++) {
				ipv6_table = blobmsg_open_table(&b, "");
				blobmsg_add_string(&b, "addr", cache->ipv6_addr[j]);
				blobmsg_close_table(&b, ipv6_table);
			}
			blobmsg_close_array(&b, ipv6_array);
		}

		if (cache->cname_cnt > 0) {
			cname_array = blobmsg_open_array(&b, "cname");
			for (size_t j = 0; j < cache->cname_cnt; j++) {
				cname_table = blobmsg_open_table(&b, "");
				blobmsg_add_string(&b, "name", cache->cname[j]);
				blobmsg_close_table(&b, cname_table);
			}
			blobmsg_close_array(&b, cname_array);
		}

		blobmsg_close_table(&b, cache_table);
	}

	blobmsg_close_array(&b, cache_array);

	int i;
	cache_array = blobmsg_open_array(&b, "url_list");
	struct profile *p = NULL;

	// Iterate over the global profile list
	// so, we loop over all filters of all profiles
	// to cover all possible urls
	list_for_each_entry(p, &profiles, list) {
		struct urlfilter *filter = NULL;

		list_for_each_entry(filter, &p->filters, list) {
			// Check if the filter is enabled
			if (!filter->enable) {
				continue;
			}

			print_url_list(&filter->filter_text, &b);

			if (filter->bundle) {
				print_url_list(&filter->bundle->custom_url, &b);
				print_url_list(&filter->bundle->download_url, &b);
			}
		}
	}
	blobmsg_close_array(&b, cache_array);

	cache_array = blobmsg_open_array(&b, "dns_server");
	for (i = 0; i < dns_server_cnt; i++) {
		cache_table = blobmsg_open_table(&b, "");
		blobmsg_add_string(&b, "server", dns_server_ip[i]);
		blobmsg_close_table(&b, cache_table);
	}
	blobmsg_close_array(&b, cache_array);

	ubus_send_reply(context, req, b.head);

	return 0;
}

static int show_config(struct ubus_context *context, struct ubus_object *obj,
			struct ubus_request_data *req, const char *method,
			struct blob_attr *msg)
{
	void *profile_array = NULL;
	struct blob_buf b = {0};
	memset(&b, 0, sizeof(struct blob_buf));
	blob_buf_init(&b, 0);

	profile_array = blobmsg_open_array(&b, "profiles");
	struct profile *prof = NULL;
	list_for_each_entry(prof, &profiles, list) {
		print_profile(prof, &b);
	}

	blobmsg_close_array(&b, profile_array);

	ubus_send_reply(context, req, b.head);
	return 0;
}

static struct ubus_method urlfilter_methods[] = {
	UBUS_METHOD_NOARG("dump", dump_dns_cache),
	UBUS_METHOD_NOARG("show_config", show_config),
};

static struct ubus_object_type urlfilter_type =
	UBUS_OBJECT_TYPE("urlfilter", urlfilter_methods);

static struct ubus_object urlfilter_object = {
	.name = "urlfilter",
	.type = &urlfilter_type,
	.methods = urlfilter_methods,
	.n_methods = ARRAY_SIZE(urlfilter_methods),
};

static int urlfilter_publish_object(struct ubus_context *context)
{
	int ret;

	ret = ubus_add_object(context, &urlfilter_object);
	if (ret)
		fprintf(stderr, "Failed to add 'urlfilter' ubus object: %s\n",
				ubus_strerror(ret));

	return ret;
}

int urlfilter_ubus_init(const char *path)
{
        uloop_init();
        ubus_path = path;

        ctx = ubus_connect(path);
        if (!ctx) {
		syslog(LOG_ERR, "Failed to connect to ubus");
                return -EIO;
	}
	
	ubus_add_uloop(ctx);

	if (urlfilter_publish_object(ctx)) {
		syslog(LOG_ERR, "Failed to publish ubus object");
		return -1;
	}

        return 0;
}

void urlfilter_ubus_done(void)
{
        ubus_free(ctx);
}

